MODULE Main;

(* ---------------------------------------------------------------------------
 * (C) 2008 - 2010 by Alexander Iljin
 * --------------------------------------------------------------------------- *)

IMPORT
   Win:=Windows, Sci:=Scintilla, Npp, OPS, Project, Tools, Compiler, Settings,
   Str;

(* ---------------------------------------------------------------------------
 * This is the Oberon IDE plugin for Notepad++ (written in XDS Oberon).
 *
 * For the free XDS Modula-2/Oberon-2 compiler go to:
 *   http://www.excelsior-usa.com/xdsx86win.html
 *   (6.6 Mb to download, 17 Mb installed);
 * Oberon-2 is object-oriented programming language, both powerful and simple.
 * The full language report is only 20 pages long:
 *   http://europrog.ru/paper/oberon-2.pdf
 * It was created by prof. N.Wirth, author of Pascal and Modula. Some of his
 * publications:
 *   Programming in Oberon: http://europrog.ru/book/obnw2004e.pdf
 *   Compiler Construction: http://europrog.ru/book/ccnw2005e.pdf
 *   Project Oberon  The Design of an Operating System and Compiler:
 *     http://europrog.ru/book/ponw2005e.pdf
 * --------------------------------------------------------------------------- *)

CONST
   PluginName* = 'Obide';

   (* Menu items *)
   AutocompletionStr*  = 'Autocompletion...';
   CodeHintStr*        = 'Code Hint';
   GotoDeclarationStr* = 'Go to Declaration';
   GoBackStr*          = 'Go Back';
   GotoCompErrStr*     = 'Go to Compiler Error';
   CompileStr*         = 'Compile';
   AboutStr*           = 'About...';

   CRLF = ''+0DX+0AX;
   AboutMsg = 'This plugin will turn your Notepad++ into Oberon IDE.'+CRLF
      +CRLF
      +'Created by Alexander Iljin (Amadeus IT Solutions) using XDS Oberon, 30 Mar 2008 - 04 Jul 2010.'+CRLF
      +'Contact e-mail: AlexIljin@users.SourceForge.net';

TYPE
   JumpPos = POINTER TO RECORD
      file: Tools.Path;
      pos, anchor: LONGINT;
      next: JumpPos;
   END;

VAR
   jumps    : JumpPos;
   lastError: Compiler.Error;

PROCEDURE About* (owner: Win.HWND; title: ARRAY OF CHAR);
(* Show info about this plugin. *)
BEGIN
   Win.MessageBox (owner, AboutMsg, title, Win.MB_OK)
END About;

PROCEDURE OpenModule (VAR name: Project.Name): BOOLEAN;
(* Open module file in the Notepad++. The name is the module name or a file
 * name without path. Return TRUE if the file is found and opened. *)
VAR dir: Tools.Path;
BEGIN
   Npp.GetCurrentDir (dir);
   RETURN Tools.OpenLoc (Tools.GetModuleLocator (name, dir));
END OpenModule;

PROCEDURE CompileCurrentFile (): Compiler.Error;
VAR fn: Tools.Path;
BEGIN
   Npp.GetFullCurrentPath (fn);
   RETURN Compiler.Compile (fn);
END CompileCurrentFile;

PROCEDURE GetModuleLocator (VAR modName: Project.Name): Project.Locator;
VAR dir: Tools.Path;
BEGIN
   Npp.GetCurrentDir (dir);
   RETURN Tools.GetModuleLocator (modName, dir);
END GetModuleLocator;

PROCEDURE GetImportedModule (im: Project.ImportedModule): Project.Module;
(* Return the parsed Module object referenced by the im in the IMPORT section.
 * You may use this function to make sure an imported module is parsed and
 * up to date. Return NIL if the module file could not be found or parsed. *)
VAR
   nm : Project.Name;
   res: Project.Module;
BEGIN
   res := NIL;
   IF im # NIL THEN
      nm := im.modName;
      res := Project.GetModule (nm, GetModuleLocator (nm));
   END;
   RETURN res
END GetImportedModule;

PROCEDURE IdentifyContext (sc: Sci.Handle; pos: LONGINT;
(* OUT *) VAR id: Tools.Ident; VAR lst: Project.List; VAR itm: Project.Item;
   VAR m: Project.Module);
(* id returns the current identifier's part from its beginning upto the caret.
 * lst returns the list of possible identifiers at caret (id then may contain
 * a partially entered ident).
 * itm returns the identified project structure item (Project.Module,
 * Project.Const, etc.) if the current identifier at cursor is the name of
 * such an item (it is searched for in the lst, but identified by the full
 * name in text, not by partial name returned in id).
 * m returns the current Module in sc. *)

   PROCEDURE IdentToItm (pos: LONGINT);
   (* Set 'itm' to contain identifier at 'pos', if any. The 'itm' is looked
    * for in the current 'lst'. *)
   VAR id2: Tools.Ident;
   BEGIN
      id2 := Tools.ReadIdentFrom (sc, pos);
      IF id2 # NIL THEN
         itm := Project.FindItem (lst, id2.name);
      ELSE
         itm := NIL;
      END;
   END IdentToItm;

BEGIN
   id := NIL; lst := NIL; itm := NIL; m := NIL;
   IF (0 <= pos) & (pos <= Sci.GetTextLength (sc)) THEN
      Tools.ParseScintilla (sc, m);
      IF m # NIL THEN
         Project.ListSubitems (m, m, GetModuleLocator, pos, lst);
         Project.RemoveInheritedMethods (lst);
         IF lst # NIL THEN
            id := Tools.GetQualidentAt (sc, pos - 1);
            IF id # NIL THEN (* there is an identifier chain *)
               itm := Project.FindItem (lst, id.name);
               WHILE (id.next # NIL) & (itm # NIL) DO
                  Project.ListSubitems (itm, m, GetModuleLocator, 0, lst);
                  Project.RemoveInheritedMethods (lst);
                  IF lst # NIL THEN
                     id := id.next;
                     itm := Project.FindItem (lst, id.name);
                  ELSE
                     itm := NIL;
                  END;
               END;
               IF id.next = NIL THEN
                  IdentToItm (id.pos);
               ELSE
                  itm := NIL;
                  lst := NIL;
               END;
            ELSE (* there still might be a single identifier -> itm *)
               IdentToItm (pos);
            END;
         END;
      END;
   END;
END IdentifyContext;

PROCEDURE Min (i1, i2: LONGINT): LONGINT;
BEGIN
   IF i1 < i2 THEN
      RETURN i1
   ELSE
      RETURN i2
   END;
END Min;

PROCEDURE ['C'] Autocompletion* ();
(* Show autocompletion list. *)
VAR
   sc : Sci.Handle;
   lst: Project.List;
   itm: Project.Item;
   id : Tools.Ident;
   mod: Project.Module;
   s  : POINTER TO ARRAY OF CHAR;
   i  : LONGINT;
BEGIN
   sc := Npp.GetCurrentScintilla ();
   IdentifyContext (sc, Sci.GetCurrentPos (sc), id, lst, itm, mod);
   IF lst # NIL THEN
      i := Sci.LinesOnScreen (sc) DIV 2; (* maximum sensible size *)
      Sci.AutoCSetMaxHeight (sc, Min (Settings.autoCMaxHeight, i));
      NEW (s, Project.NamesLength (lst));
      Project.NamesToString (lst, s^);
      IF id # NIL THEN
         WHILE id.next # NIL DO
            id := id.next;
         END;
         i := id.Length ();
      ELSE
         i := 0;
      END;
      Sci.AutoCShow (sc, i, s^);
   END;
END Autocompletion;

PROCEDURE GetImportedModuleHint (moduleName, moduleFile: ARRAY OF CHAR;
   VAR res: ARRAY OF CHAR);
CONST
   ImportedModuleStr = 'Imported module ';
   FileNotFoundStr = '' + 0AX + 'File not found (' + Settings.GlobalSection + '.' + Settings.LookupKey + '=';
   FileFoundStr = '' + 0AX + 'File: ';
VAR
   i, c: LONGINT;
BEGIN
   COPY (ImportedModuleStr, res);
   Str.AppendC (res, moduleName);
   IF moduleFile [0] # 0X THEN
      Str.AppendC (res, FileFoundStr);
      Str.AppendC (res, moduleFile);
   ELSE
      Str.AppendC (res, FileNotFoundStr);
      IF Settings.lookup # NIL THEN
         Str.Append (res, Settings.lookup [0]^);
         i := 1;
         c := LEN (Settings.lookup^);
         WHILE i < c DO
            Str.AppendC (res, Settings.LookupSeparator);
            Str.Append (res, Settings.lookup [i]^);
            INC (i);
         END;
         Str.AppendC (res, ')');
      END;
   END;
END GetImportedModuleHint;

PROCEDURE CodeHintAt (VAR sc: Sci.Handle; pos: LONGINT);
(* Show a hint about the current identifier. *)
VAR
   lst: Project.List;
   itm: Project.Item;
   id : Tools.Ident;
   mod: Project.Module;
   s  : OPS.String;
   i  : LONGINT;
BEGIN
   IdentifyContext (sc, pos, id, lst, itm, mod);
   lst := NIL;
   IF itm # NIL THEN
      WITH itm: Project.Const DO
         COPY (itm.val, s);
      | itm: Project.Var DO
         Project.TypeToDeclString (itm.type, mod, itm.mod, s);
      | itm: Project.TypeItem DO
         Project.TypeToDeclString (itm.type, mod, itm.mod, s);
      | itm: Project.Proc DO
         COPY (itm.val, s);
         IF s = '' THEN
            s := '()';
         END;
      | itm: Project.ImportedModule DO
         mod := GetImportedModule (itm);
         IF (mod # NIL) & (mod.loc # NIL) & (mod.loc IS Tools.Locator) THEN
            GetImportedModuleHint (itm.modName, mod.loc (Tools.Locator).path, s);
         ELSE
            GetImportedModuleHint (itm.modName, '', s);
         END;
      ELSE
         s := '';
      END;
      IF s # '' THEN
         i := pos;
         IF id # NIL THEN
            DEC (i, id.Length ());
         END;
         Sci.CallTipShow (sc, i, s);
      END;
   END;
END CodeHintAt;

PROCEDURE ['C'] CodeHint* ();
(* Show a hint about the current identifier. *)
VAR
   sc: Sci.Handle;
BEGIN
   sc := Npp.GetCurrentScintilla ();
   CodeHintAt (sc, Sci.GetCurrentPos (sc));
END CodeHint;

PROCEDURE GetCurrentPos (sc: Sci.Handle): JumpPos;
(* Return the current file and position as a jump target. *)
VAR res: JumpPos;
BEGIN
   NEW (res);
   Npp.GetFullCurrentPath (res.file);
   res.pos := Sci.GetCurrentPos (sc);
   res.anchor := Sci.GetAnchor (sc);
   RETURN res
END GetCurrentPos;

PROCEDURE SamePos (pos1, pos2: JumpPos): BOOLEAN;
BEGIN
   RETURN (pos1.file = pos2.file) & (pos1.pos = pos2.pos);
END SamePos;

PROCEDURE RegisterJump (jmp: JumpPos; sc: Sci.Handle);
(* Put the jmp on top of the jump stack, unless the current position is the
 * same as the jump target. *)
VAR curr: JumpPos;
BEGIN
   curr := GetCurrentPos (sc);
   IF ~SamePos (jmp, curr) THEN
      jmp.next := jumps;
      jumps := jmp;
   END;
END RegisterJump;

PROCEDURE ['C'] GotoDeclaration* ();
(* Jump the caret to the declaration place of the current identifier. Open the
 * necessary module file, if required. *)
VAR
   sc : Sci.Handle;
   lst: Project.List;
   itm: Project.Item;
   id : Tools.Ident;
   mod: Project.Module;
   jmp: JumpPos;

   PROCEDURE ShowModuleNotFoundHint (itm: Project.ImportedModule; id: Tools.Ident);
   VAR
      s: ARRAY 1024 OF CHAR;
      i: LONGINT;
   BEGIN
      GetImportedModuleHint (itm.modName, '', s);
      i := Sci.GetCurrentPos (sc);
      IF id # NIL THEN
         DEC (i, id.Length ());
      END;
      Sci.CallTipShow (sc, i, s);
   END ShowModuleNotFoundHint;

BEGIN
   sc := Npp.GetCurrentScintilla ();
   IdentifyContext (sc, Sci.GetCurrentPos (sc), id, lst, itm, mod);
   IF itm # NIL THEN
      IF itm IS Project.ImportedModule THEN
         (* special case: if the user asks to jump to the module name declaration,
          * open the module file, not the IMPORT section with the identifier.
          * Only jump to the IMPORT section if the module file is not found. *)
         mod := GetImportedModule (itm (Project.ImportedModule));
         IF mod # NIL THEN
            itm := mod; (* jump to module name *)
         ELSE
            ShowModuleNotFoundHint (itm (Project.ImportedModule), id);
         END;
      ELSE
         mod := itm.mod;
      END;
   END;
   IF (itm # NIL) & (mod # NIL) THEN
      jmp := GetCurrentPos (sc);
      IF Tools.OpenModule (mod) THEN
         (* it can be just the current Scintilla buffer, not a file *)
      END;
      Sci.GotoPos (sc, itm.pos);
      RegisterJump (jmp, sc);
   END;
END GotoDeclaration;

PROCEDURE ['C'] GoBack* ();
(* Jump back to the previous position before the last call to
 * GotoDeclaration. *)
VAR sc: Sci.Handle;
BEGIN
   IF jumps # NIL THEN
      IF Tools.OpenFile (jumps.file) THEN;
         sc := Npp.GetCurrentScintilla ();
         Sci.SetCurrentPos (sc, jumps.pos);
         Sci.SetAnchor (sc, jumps.anchor);
         Sci.ScrollCaret (sc);
      END;
      jumps := jumps.next;
   END;
END GoBack;

PROCEDURE GotoLastError ();
CONST
   SuccessStr = 'Compiled successfully.';
VAR
   sc : Sci.Handle;
   pos: LONGINT;
   mod: Project.Name;
   i  : INTEGER;
   msg: Compiler.Str;
   jmp: JumpPos;
BEGIN
   sc := Npp.GetCurrentScintilla ();
   pos := Sci.GetCurrentPos (sc);
   IF lastError # NIL THEN
      IF (lastError.file # '') & (lastError.line > 0) & (lastError.col > 0) THEN
         i := 0;
         IF Tools.IdentFirstChar (lastError.file [i]) THEN
            REPEAT
               mod [i] := lastError.file [i];
               INC (i);
            UNTIL ~Tools.IdentChar (lastError.file [i]) OR (i >= LEN (mod) - 1);
         END;
         mod [i] := 0X;
         jmp := GetCurrentPos (sc);
         IF OpenModule (mod) THEN
            pos := Sci.PositionFromLine (sc, lastError.line - 1) + lastError.col - 1;
            Sci.SetCurrentPos (sc, pos);
            Sci.SetAnchor (sc, pos);
            Sci.ScrollCaret (sc);
            RegisterJump (jmp, sc);
         END;
      END;
      msg := lastError.msg;
   ELSE
      msg := SuccessStr;
   END;
   Sci.CallTipShow (sc, pos, msg);
END GotoLastError;

PROCEDURE ['C'] Compile* ();
(* Compile the current project and show the error message, if any. *)
BEGIN
   lastError := CompileCurrentFile ();
   GotoLastError;
END Compile;

PROCEDURE ['C'] GotoCompErr* ();
(* Jump to the latest compiler error position. *)
BEGIN
   GotoLastError;
END GotoCompErr;

PROCEDURE NppReady* (scintillaMain, scintillaSecond: Sci.Handle);
(* This procedure is executed when Notepad++ has finished its startup. *)
CONST IniFileName = PluginName + '.ini';
VAR
   iniFile: ARRAY Win.MAX_PATH OF CHAR;
   autoCChars: ARRAY 24 OF CHAR;
BEGIN
   (* Load settings first, then apply them to Scintilla *)
   Npp.GetPluginConfigDir (iniFile);
   Str.AppendC (iniFile, '\');
   Str.AppendC (iniFile, IniFileName);
   Settings.LoadFromFile (iniFile);
   (* When a jump is performed and the caret goes off screen, scroll it to the
    * center *)
   Sci.SetYCaretPolicy (scintillaMain, Sci.caretJumps + Sci.caretEven, 0);
   Sci.SetYCaretPolicy (scintillaSecond, Sci.caretJumps + Sci.caretEven, 0);
   (* Allow scrolling one page past end of file *)
   Sci.SetEndAtLastLine (scintillaMain, FALSE);
   Sci.SetEndAtLastLine (scintillaSecond, FALSE);
   (* Replace the rest of the curret ident when selection was made *)
   Sci.AutoCSetDropRestOfWord (scintillaMain, TRUE);
   Sci.AutoCSetDropRestOfWord (scintillaSecond, TRUE);
   (* Don't close the autocompletion list unless the ident was erased *)
   Sci.AutoCSetCancelAtStart (scintillaMain, FALSE);
   Sci.AutoCSetCancelAtStart (scintillaSecond, FALSE);
   (* Typing the following characters confirms autocompletion selection *)
   autoCChars := ' ([{}]).:;,+-*/|<>=#^&' + 09X (* 09X = Tab character *);
   Sci.AutoCSetFillUps (scintillaMain, autoCChars);
   Sci.AutoCSetFillUps (scintillaSecond, autoCChars);
   (* Set length of the delay before the Code Hint is displayed under mouse *)
   Sci.SetMouseDwellTime (scintillaMain, Settings.mouseHintDelay);
   Sci.SetMouseDwellTime (scintillaSecond, Settings.mouseHintDelay);
END NppReady;

PROCEDURE ScintillaNotify* (VAR sci: Sci.Handle; VAR note: Sci.SCNotification);
BEGIN
   CASE note.nmhdr.code OF
   |  Sci.noteDwellStart:
      Sci.CallTipCancel (sci);
      CodeHintAt (sci, note.position);
   ELSE
   END;
END ScintillaNotify;

END Main.
