<*+main*> (* This marks the main module of a program or library.               *)
<*heaplimit="2000000"*> (* Maximum heap size should be set in the main module,
because the changes do not take effect until the main module is recompiled.    *)

MODULE ObideU;

(* ---------------------------------------------------------------------------
 * (C) 2008 - 2010 by Alexander Iljin
 * --------------------------------------------------------------------------- *)

IMPORT
   Win:=Windows, NotepadPP:=NotepadPPU, Npp, StrU, ObideUVer, Main;--,Debug;

(* ---------------------------------------------------------------------------
 * This is the Oberon IDE plugin for Notepad++ (written in XDS Oberon).
 *
 * For the free XDS Modula-2/Oberon-2 compiler go to:
 *   http://www.excelsior-usa.com/xdsx86win.html
 *   (6.6 Mb to download, 17 Mb installed);
 * Oberon-2 is object-oriented programming language, both powerful and simple.
 * The full language report is only 20 pages long:
 *   http://europrog.ru/paper/oberon-2.pdf
 * It was created by prof. N.Wirth, author of Pascal and Modula. Some of his
 * publications:
 *   Programming in Oberon: http://europrog.ru/book/obnw2004e.pdf
 *   Compiler Construction: http://europrog.ru/book/ccnw2005e.pdf
 *   Project Oberon  The Design of an Operating System and Compiler:
 *     http://europrog.ru/book/ponw2005e.pdf
 * --------------------------------------------------------------------------- *)

PROCEDURE ['C'] About ();
(* Show info about this plugin. *)
BEGIN
   Main.About (NotepadPP.handle, Main.PluginName + ' ' + ObideUVer.String);
END About;

PROCEDURE NppReady ();
(* This procedure is executed when Notepad++ has finished its startup. *)
BEGIN
   Main.NppReady (NotepadPP.scintillaMain, NotepadPP.scintillaSecond);
END NppReady;

PROCEDURE UnicodeToAcp (VAR in: ARRAY OF NotepadPP.Char; VAR out: ARRAY OF CHAR);
VAR cnt: LONGINT;
BEGIN
   cnt := Win.WideCharToMultiByte (Win.CP_ACP, Win.WC_NO_BEST_FIT_CHARS, in [0], -1, out [0], LEN (out), NIL, NIL);
   IF cnt = 0 THEN
      out [0] := 0X;
      Win.MessageBox (NotepadPP.handle, 'Error converting Unicode path to ANSI code page. Sorry, but Unicode paths are not fully supported by the Obide plugin.', Main.PluginName, Win.MB_OK);
   END;
END UnicodeToAcp;

PROCEDURE AcpToUnicode (VAR in: ARRAY OF CHAR; VAR out: ARRAY OF NotepadPP.Char);
VAR cnt: LONGINT;
BEGIN
   cnt := Win.MultiByteToWideChar (Win.CP_ACP, Win.MULTIBYTE_SET {}, in [0], -1, out [0], LEN (out));
   IF cnt = 0 THEN
      out [0] := StrU.Null;
      Win.MessageBox (NotepadPP.handle, 'Error converting ANSI path to Unicode. Sorry, but Unicode paths are not fully supported by the Obide plugin.', Main.PluginName, Win.MB_OK);
   END;
END AcpToUnicode;

PROCEDURE GetFullCurrentPath* (VAR res: ARRAY OF CHAR);
VAR tmp: ARRAY 1024 OF NotepadPP.Char;
BEGIN
   NotepadPP.GetFullCurrentPath (tmp);
   UnicodeToAcp (tmp, res);
END GetFullCurrentPath;

PROCEDURE GetCurrentDir* (VAR res: ARRAY OF CHAR);
VAR tmp: ARRAY 1024 OF NotepadPP.Char;
BEGIN
   NotepadPP.GetCurrentDir (tmp);
   UnicodeToAcp (tmp, res);
END GetCurrentDir;

PROCEDURE GetPluginConfigDir (VAR res: ARRAY OF CHAR);
VAR tmp: ARRAY 1024 OF NotepadPP.Char;
BEGIN
   NotepadPP.GetPluginConfigDir (tmp);
   UnicodeToAcp (tmp, res);
END GetPluginConfigDir;

PROCEDURE OpenFile* (fname: ARRAY OF CHAR): BOOLEAN;
VAR tmp: ARRAY 1024 OF NotepadPP.Char;
BEGIN
   AcpToUnicode (fname, tmp);
   RETURN NotepadPP.OpenFile (tmp)
END OpenFile;

PROCEDURE Init ();
VAR s: NotepadPP.Shortcut;

   PROCEDURE AddMenuItem (name: ARRAY OF CHAR; func: NotepadPP.Function; initChk: BOOLEAN; shortcut: NotepadPP.Shortcut);
   VAR menuItemStr: ARRAY NotepadPP.MenuItemNameLength OF NotepadPP.Char;
   BEGIN
      StrU.Assign (name, menuItemStr);
      NotepadPP.AddMenuItem (menuItemStr, func, initChk, shortcut);
   END AddMenuItem;

BEGIN
   (* Initialize the Npp interface module *)
   Npp.GetCurrentScintilla := NotepadPP.GetCurrentScintilla;
   Npp.GetPluginConfigDir := GetPluginConfigDir;
   Npp.GetCurrentDir := GetCurrentDir;
   Npp.GetFullCurrentPath := GetFullCurrentPath;
   Npp.OpenFile := OpenFile;
   (* Initialize the NotepadPP module *)
   StrU.Assign (Main.PluginName, NotepadPP.PluginName);
   NotepadPP.onReady := NppReady;
   NEW (s);
   s.key := ' ';
   s.ctrl := TRUE;
   AddMenuItem (Main.AutocompletionStr, Main.Autocompletion, FALSE, s);
   NEW (s);
   s.key := ' ';
   s.ctrl := TRUE;
   s.alt := TRUE;
   AddMenuItem (Main.CodeHintStr, Main.CodeHint, FALSE, s);
   NEW (s);
   s.key := '9';
   s.ctrl := TRUE;
   AddMenuItem (Main.GotoDeclarationStr, Main.GotoDeclaration, FALSE, s);
   NEW (s);
   s.key := '0';
   s.ctrl := TRUE;
   AddMenuItem (Main.GoBackStr, Main.GoBack, FALSE, s);
   NEW (s);
   s.key := '-';
   s.ctrl := TRUE;
   AddMenuItem (Main.GotoCompErrStr, Main.GotoCompErr, FALSE, s);
   NEW (s);
   s.key := '1';
   s.ctrl := TRUE;
   AddMenuItem (Main.CompileStr, Main.Compile, FALSE, s);
   NotepadPP.AddMenuSeparator;
   AddMenuItem (Main.AboutStr, About, FALSE, NIL)
END Init;

BEGIN Init
END ObideU.
