<*+main*>
MODULE Test;

(** ------------------------------------------------------------------------------
 ** (C) 2008 by Alexander Iljin
 ** ------------------------------------------------------------------------------
 ** This is a bunch of tests to check the Project module.
 ** ------------------------------------------------------------------------------ *)

IMPORT
   Tests,Tools,Project,Streams,MemStream,Str,OPS;

CONST
   testName = 'Project module'; (* Name of the test *)

PROCEDURE ReadThis (VAR s: Streams.Stream; VAR str: ARRAY OF CHAR);
VAR ms: MemStream.Stream; i: LONGINT;
BEGIN
   NEW (ms); ms.Init; ms.Assign (str);
   ASSERT (ms.Open (FALSE, FALSE));
   i := 0; WHILE (i < LEN (str)) & (str [i] # 0X) DO INC (i) END;
   ms.SetSize (i);
   s := ms
END ReadThis;

PROCEDURE ReadThisC (VAR s: Streams.Stream; input: ARRAY OF CHAR);
BEGIN
   ReadThis (s, input)
END ReadThisC;

PROCEDURE ParseImport ();
VAR
   s: Streams.Stream;
   res: ARRAY 10 OF OPS.Name;
   nRes: INTEGER;
   m : Project.Module;
   ok: BOOLEAN;

   PROCEDURE AddResultStr (str: ARRAY OF CHAR);
   BEGIN
      COPY (str, res [nRes]);
      INC (nRes);
      ASSERT (nRes <= LEN (res), 60);
   END AddResultStr;

   PROCEDURE RunTest (input: ARRAY OF CHAR);
   VAR
      n  : INTEGER;
      itm: Project.Item;
   BEGIN
      ReadThis (s, input);
      Tools.ParseStream (s, m);
      ASSERT (m # NIL, 100);
      itm := m.import;
      n := 0;
      WHILE (itm # NIL) & (itm.name = res [n]) DO
         -- Out.Ln; Out.String ('read str: "'); Out.String (str);
         -- Out.String ('"; Length (str) = '); Out.Int (Str.Length (str), 0);
         INC (n);
         itm := itm.next
      END;
      -- Out.Ln;
      -- Out.String ('s.err = '); Out.Int (s.err, 0); Out.Ln;
      -- Out.String ('s.byteCnt = '); Out.Int (s.byteCnt, 0); Out.Ln;
      -- Out.String ('str = "'); Out.String (str); Out.String ('"; Length (str) = '); Out.Int (Str.Length (str), 0); Out.Ln;
      -- Out.String ('n = '); Out.Int (n, 0); Out.Ln;
      ok := (n = nRes) & (itm = NIL);
      nRes := 0;
   END RunTest;

BEGIN
   Tests.Begin ('ParseImport');
   nRes := 0;
   AddResultStr ('beta');
   AddResultStr ('chk');
   AddResultStr ('zet');
   RunTest ('MODULE X; IMPORT zet,beta:=alpha,chk;');
   Tests.Report (ok)
END ParseImport;

PROCEDURE ParseConst ();
VAR
   s: Streams.Stream;
   res: ARRAY 10 OF RECORD name: OPS.Name; val: OPS.String END;
   nRes: INTEGER;
   m : Project.Module;
   ok: BOOLEAN;

   PROCEDURE AddResult (str, val: ARRAY OF CHAR);
   BEGIN
      COPY (str, res [nRes].name);
      COPY (val, res [nRes].val);
      INC (nRes);
      ASSERT (nRes <= LEN (res), 60);
   END AddResult;

   PROCEDURE RunTest (input: ARRAY OF CHAR);
   VAR
      n: INTEGER;
      c: Project.Const;
      done: BOOLEAN;
   BEGIN
      ReadThis (s, input);
      Tools.ParseStream (s, m);
      ASSERT (m # NIL, 100);
      c := m.const; n := 0;
      done := (c = NIL) OR (c.name # res [n].name) OR (c.val # res [n].val);
      WHILE ~done DO
         INC (n);
         IF c.next # NIL THEN
            c := c.next (Project.Const);
            done := (c.name # res [n].name) OR (c.val # res [n].val);
         ELSE
            c := NIL;
            done := TRUE
         END
      END;
      ok := (n = nRes) & (c = NIL);
      nRes := 0;
   END RunTest;

BEGIN
   Tests.Begin ('ParseConst');
   nRes := 0;
   AddResult ('beta', 'M.alpha+15');
   AddResult ('chk', '"hello"+0DX');
   AddResult ('zet', '1.0E+4');
   RunTest ('MODULE X; CONST zet = 1.0E+4; beta=(*comment*)M.alpha+15 ;chk="hello" + 0DX');
   Tests.Report (ok)
END ParseConst;

PROCEDURE ParseToString ();
VAR
   s  : Streams.Stream;
   m  : Project.Module;
   lst: Project.List;
   str: Str.PStr;
   ok: BOOLEAN;
BEGIN
   Tests.Begin ('ParseToString');
   ReadThisC (s, '-- head comment'+0DX+0AX
      +'MODULE X; IMPORT Win:=Windows,Module1;'+0DX+0AX
      +'CONST zet = <*compiler="directive"*> 1.0E+4; beta=(*comment*)--M.alpha+15 ;'+0DX+0AX
      +'M.alpha+15 ;chk="hello" + 0DX;');
   Tools.ParseStream (s, m);
   IF m # NIL THEN
      Project.ListSubitems (m, m, NIL, 0, lst);
      NEW (str, Project.NamesLength (lst));
      Project.NamesToString (lst, str^);
      ok := str^ = 'Module1 Win beta chk zet';
   ELSE
      ok := FALSE
   END;
   Tests.Report (ok)
END ParseToString;

PROCEDURE ParseProc ();
VAR
   s  : Streams.Stream;
   m  : Project.Module;
   lst: Project.List;
   str: Str.PStr;
   ok: BOOLEAN;
BEGIN
   Tests.Begin ('ParseProc');
   ReadThisC (s, 'MODULE X;'+0DX+0AX
      +'PROCEDURE A; BEGIN END END A;'+0DX+0AX
      +'PROCEDURE B; BEGIN END B;'+0DX+0AX
      +'END X.');
   Tools.ParseStream (s, m);
   IF m # NIL THEN
      Project.ListSubitems (m, m, NIL, 0, lst);
      NEW (str, Project.NamesLength (lst));
      Project.NamesToString (lst, str^);
      ok := str^ = 'A B';
   ELSE
      ok := FALSE
   END;
   Tests.Report (ok)
END ParseProc;

PROCEDURE Do;
BEGIN
   ParseImport;
   ParseConst;
   ParseToString;
   ParseProc;
END Do;

BEGIN
   Tests.Begin (testName); Do; Tests.End;
END Test.
